
#if defined(_MSC_VER) && _MSC_VER >= 1200
#pragma warning(push, 3)
#endif

/* disable or reduce the frequency of...
 *   C4057: indirection to slightly different base types
 *   C4075: slight indirection changes (unsigned short* vs short[])
 *   C4100: unreferenced formal parameter
 *   C4127: conditional expression is constant
 *   C4163: '_rotl64' : not available as an intrinsic function
 *   C4201: nonstandard extension nameless struct/unions
 *   C4244: int to char/short - precision loss
 *   C4514: unreferenced inline function removed
 */
#pragma warning(disable: 4100 4127 4163 4201 4514; once: 4057 4075 4244)

/* Ignore Microsoft's interpretation of secure development
 * and the POSIX string handling API
 */
#if defined(_MSC_VER) && _MSC_VER >= 1400
#ifndef _CRT_SECURE_NO_DEPRECATE
#define _CRT_SECURE_NO_DEPRECATE
#endif
#pragma warning(disable: 4996)
#endif

/* Has windows.h already been included?  If so, our preferences don't matter,
 * but we will still need the winsock things no matter what was included.
 * If not, include a restricted set of windows headers to our tastes.
 */
#ifndef _WINDOWS_
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#ifndef _WIN32_WINNT

/* Restrict the server to a subset of Windows NT 4.0 header files by default
 */
#define _WIN32_WINNT 0x0400
#endif
#ifndef NOUSER
#define NOUSER
#endif
#ifndef NOMCX
#define NOMCX
#endif
#ifndef NOIME
#define NOIME
#endif

#include <windows.h>
/*
 * Add a _very_few_ declarations missing from the restricted set of headers
 * (If this list becomes extensive, re-enable the required headers above!)
 * winsock headers were excluded by WIN32_LEAN_AND_MEAN, so include them now
 */
#define SW_HIDE             0
#ifndef _WIN32_WCE
#include <winsock2.h>
#include <mswsock.h>
#include <ws2tcpip.h>
#else
#include <winsock.h>
#endif
#endif /* !_WINDOWS_ */

#include <jni.h>
#include <io.h>

#define IS_NULL(obj) ((obj) == NULL)

static JavaVM     *global_jvm = NULL;


void dbprintf(char *format, ...)
{
    va_list args;
    char buffer[4096];

    va_start(args, format);
    _vsnprintf(buffer, 4096, format, args);
    va_end(args);
    OutputDebugString(buffer);
}

/* field id for jint 'fd' in java.io.FileDescriptor */
jfieldID IO_fd_fdID = NULL;

/* field id for jlong 'handle' in java.io.FileDescriptor */
jfieldID IO_handle_fdID = NULL;

jfieldID fis_fd = NULL; /* id for jobject 'fd' in java.io.FileInputStream */

/**************************************************************
 * static methods to store field IDs in initializers
 */

static void
FileDescriptor_initIDs(JNIEnv *env, jclass fdClass)
{
    IO_fd_fdID = (*env)->GetFieldID(env, fdClass, "fd", "I");
    IO_handle_fdID = (*env)->GetFieldID(env, fdClass, "handle", "J");
}


static void
FileInputStream_initIDs(JNIEnv *env, jclass fdClass) {
    fis_fd = (*env)->GetFieldID(env, fdClass, "fd", "Ljava/io/FileDescriptor;");
}

/**
 * Throw a Java exception by name. Similar to SignalError.
 */
static void
JNU_ThrowByName(JNIEnv *env, const char *name, const char *msg)
{
    jclass cls = (*env)->FindClass(env, name);

    if (cls != 0) /* Otherwise an exception has already been thrown */
    (*env)->ThrowNew(env, cls, msg);
}

/* JNU_Throw common exceptions */

static void
JNU_ThrowNullPointerException(JNIEnv *env, const char *msg)
{
    JNU_ThrowByName(env, "java/lang/NullPointerException", msg);
}


static void
JNU_ThrowOutOfMemoryError(JNIEnv *env, const char *msg)
{
    JNU_ThrowByName(env, "java/lang/OutOfMemoryError", msg);
}

static jstring OS_syserror(JNIEnv *env, jint err)
{
    jstring str;
    void *buf;

    if (!FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                       FORMAT_MESSAGE_FROM_SYSTEM |
                       FORMAT_MESSAGE_IGNORE_INSERTS,
                       NULL,
                       err,
                       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                       (LPTSTR)&buf,
                       0,
                       NULL)) {
        str = (*env)->NewStringUTF(env, "Unknown Error");
    }
    else {
        str = (*env)->NewStringUTF(env, (const char *)buf);
        LocalFree(buf);
    }
    return str;
}


static jobject
JNU_NewObjectByName(JNIEnv *env, const char *class_name,
            const char *constructor_sig, ...)
{
    jobject obj = NULL;

    jclass cls = 0;
    jmethodID cls_initMID;
    va_list args;

    if ((*env)->EnsureLocalCapacity(env, 2) < 0)
        goto done;

    cls = (*env)->FindClass(env, class_name);
    if (cls == 0) {
    goto done;
    }
    cls_initMID  = (*env)->GetMethodID(env, cls,
                       "<init>", constructor_sig);
    if (cls_initMID == NULL) {
    goto done;
    }
    va_start(args, constructor_sig);
    obj = (*env)->NewObjectV(env, cls, cls_initMID, args);
    va_end(args);

 done:
    (*env)->DeleteLocalRef(env, cls);
    return obj;
}


/* Throw an exception by name, using the string returned by
 * JVM_LastErrorString for the detail string.  If the last-error
 * string is NULL, use the given default detail string.
 */
static void
JNU_ThrowByNameWithLastError(JNIEnv *env, const char *name,
                 const char *defaultDetail)
{

    int thrown = 0;
    DWORD n = GetLastError();
    if (n != ERROR_SUCCESS) {
        jstring s = OS_syserror(env, n);
        if (s != NULL) {
            jobject x = JNU_NewObjectByName(env, name,
                                    "(Ljava/lang/String;)V", s);
            if (x != NULL) {
            (*env)->Throw(env, x);
            thrown = 1;
            }
        }
    }
    if (!thrown) {
        JNU_ThrowByName(env, name, defaultDetail);
    }
}

static void
JNU_ThrowIOExceptionWithLastError(JNIEnv *env, const char *defaultDetail)
{
    JNU_ThrowByNameWithLastError(env, "java/io/IOException", defaultDetail);
}

/*
 * Macros to use the right data type for file descriptors
 */
#define FD jlong

/*
 * Macros to set/get fd from the java.io.FileDescriptor.
 */
#define SET_FD(this, fd, fid) (*env)->SetLongField(env, \
          (*env)->GetObjectField(env, (this), (fid)),IO_handle_fdID, (fd))
#define GET_FD(this, fid) (*env)->GetLongField(env, \
          (*env)->GetObjectField(env, (this), (fid)), IO_handle_fdID)
/*
 * Macros to set/get fd when inside java.io.FileDescriptor
 */
#define THIS_FD(obj) (*env)->GetLongField(env, obj, IO_handle_fdID)


#define JVM_IO_ERR  (-1)
#define JVM_IO_INTR (-2)

/* IO helper functions */

int
readSingle(JNIEnv *env, jobject this, jfieldID fid) {
    int nread;
    char ret;
    FD fd = GET_FD(this, fid);

    if (!ReadFile((HANDLE)fd, &ret, 1, &nread, NULL))
        nread = JVM_IO_ERR;
    if (nread == 0) { /* EOF */
        return -1;
    } else if (nread == -1) { /* error */
        JNU_ThrowIOExceptionWithLastError(env, "Read error");
    } else if (nread == JVM_IO_INTR) {
        JNU_ThrowByName(env, "java/io/InterruptedIOException", 0);
    }
    return ret & 0xFF;
}

/* The maximum size of a stack-allocated buffer.
 */
#define BUF_SIZE 8192


int
readBytes(JNIEnv *env, jobject this, jbyteArray bytes,
      jint off, jint len, jfieldID fid)
{
    int nread, datalen, rd = 0;
    char stackBuf[BUF_SIZE];
    char *buf = 0;
    FD fd;
	HANDLE h;

    if (IS_NULL(bytes)) {
        JNU_ThrowNullPointerException(env, 0);
        return -1;
    }
    datalen = (*env)->GetArrayLength(env, bytes);
    if ((off < 0) || (off > datalen) ||
        (len < 0) || ((off + len) > datalen) || ((off + len) < 0)) {
        JNU_ThrowByName(env, "java/lang/IndexOutOfBoundsException", 0);
        return -1;
    }

    if (len == 0) {
        return 0;
    } else if (len > BUF_SIZE) {
        buf = malloc(len);
        if (buf == 0) {
            JNU_ThrowOutOfMemoryError(env, 0);
            return 0;
        }
    } else {
        buf = stackBuf;
    }

    fd = GET_FD(this, fid);
	h = (HANDLE)fd;
    if (!ReadFile(h, buf, len, &nread, NULL))
        nread = JVM_IO_ERR;
    if (nread > 0) {
        (*env)->SetByteArrayRegion(env, bytes, off, nread, (jbyte *)buf);
    } else if (nread == JVM_IO_ERR) {
		int err = GetLastError();
        JNU_ThrowIOExceptionWithLastError(env, "Read error");
		dbprintf("readBytes error=%d %p readed %d", err, h, nread);
    } else if (nread == JVM_IO_INTR) { /* EOF */
        JNU_ThrowByName(env, "java/io/InterruptedIOException", 0);
    } else { /* EOF */
        nread = -1;
    }

    if (buf != stackBuf) {
        free(buf);
    }
    return nread;
}


JNIEXPORT jint JNICALL
Java_java_io_FileInputStream_readBytes(JNIEnv *env, jobject this,
                                       jbyteArray bytes,
                                       jint off, jint len)
{
    return readBytes(env, this, bytes, off, len, fis_fd);
}


/* Called by the JVM when APR_JAVA is loaded */
JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *vm, void *reserved)
{
    JNIEnv *env;

    if ((*vm)->GetEnv(vm, (void **)&env, JNI_VERSION_1_4)) {
        return JNI_ERR;
    }
    global_jvm = vm;

    return  JNI_VERSION_1_4;
}


JNIEXPORT void JNICALL
Java_JvmWrapper_initFileDescriptor(JNIEnv *env, jobject this,
                                     jclass fdClass)
{
    FileDescriptor_initIDs(env, fdClass);
}

JNIEXPORT void JNICALL
Java_JvmWrapper_initFileInputStream(JNIEnv *env, jobject this,
                                     jclass fdClass)
{
    FileInputStream_initIDs(env, fdClass);
}


static JNINativeMethod nativeMethods[] = {
    {
        "readBytes", "([BII)I",
        Java_java_io_FileInputStream_readBytes
    }
};

/*
  Register Native methods returning the total number of
  native functions
*/
JNIEXPORT void JNICALL
Java_JvmWrapper_registerNatives(JNIEnv *jniEnv, jobject this, jclass bridgeClass)
{
   (*jniEnv)->RegisterNatives(jniEnv, bridgeClass, nativeMethods, 1);
}
